/* sms (c) James McKay : TI76489.C */

#define TI76489_OBJ

#include<stdio.h>
#include "sms_end.c"

#ifdef MSDOS_SYSTEM

#include <dos.h>

#ifndef __BORLANDC__
#include<i86.h>
#endif

#else

#define _disable()
#define _enable()

#endif

#include "ti76489.h"

extern word PSG[8];
extern UC fm_chip[256];
extern UC adlib_present;
extern UC all_sound_on, quit;

UC white_noise=1, last_noise_channel=0;
UC UseAdlib=1;
UC PSG_open=0;
UC Volumes[16]=
{
  0xFF,0x3F,0x3F,0x3F,0x2C,0x26,0x22,0x1D,
  0x16,0x12,0x0C,0x0A,0x07,0x05,0x02,0x00
};

#ifdef OLD_NOISE
/*** PutAdlib ****************************************/
/*** Write value V into Adlib register R.          ***/
/*****************************************************/
void PutAdlib(register UC R,register UC V)
{
  register UC J;

  outp(0x388,R);for(J=0;J<6;J++) inp(0x388);
  outp(0x389,V);for(J=0;J<35;J++) inp(0x388);
}

/*** MuteSound ***************************************/
/*** Turn all sound off.                           ***/
/*****************************************************/
void MuteSound(void)
{
  word OldPSG;
  UC x;

  if(UseAdlib)
  {
	  for(x=0;x<4;x++)
	  {
		  OldPSG=PSG[(x<<1)|1];
		  PSG[(x<<1)|1]=15;
		  Set_Vol(x);
		  PSG[(x<<1)|1]=OldPSG;
	  }
  }
  else
  {
	nosound();
  }
}

/*** InitAdlib ***************************************/
/*** Return 0 if Adlib is not found, otherwise     ***/
/*** initialize it and return 1.                   ***/
/*****************************************************/
UC InitAdlib(void)
{
  UC I,A1,A2;

  /* If Adlib is not to be used, don't initialize */
  if(!UseAdlib) return(0);

  /********* Detect if Adlib is present. *************/
  PutAdlib(0x01,0x00);    /* Delete test register    */
  PutAdlib(0x04,0x60);    /* Mask and disable timers */
  PutAdlib(0x04,0x80);    /* Reset timers            */
  A1=inp(0x388)&0xE0;     /* Read status             */
  PutAdlib(0x02,0xFF);    /* Set Timer1 to 0xFF      */
  PutAdlib(0x04,0x21);    /* Unmask and start Timer1 */
  for(I=0;I<0xC8;I++) inp(0x388);  /* Wait 80us      */
  A2=inp(0x388)&0xE0;     /* Read status             */
  PutAdlib(0x04,0x60);    /* Mask and disable timers */
  PutAdlib(0x04,0x80);    /* Reset timers            */
  if((A2!=0xC0)||A1) { UseAdlib=0;return(0); }
  /***************************************************/

  /* Allow the FM chips to control WaveForm */
  for(I=0;I<255;I++) PutAdlib(I,0x00);
  PutAdlib(0x01,0x20);
  PutAdlib(0xBD,0x20|8);

  /* Turn sound off */
  MuteSound();

  for(I=0;I<3;I++)
  {
    PutAdlib(0x20+I,0x01);
    PutAdlib(0x23+I,0x01);
    PutAdlib(0x40+I,0x18);
    PutAdlib(0x43+I,0x3F);
    PutAdlib(0x60+I,0xF0);
    PutAdlib(0x63+I,0xF0);
    PutAdlib(0x80+I,0x14);
    PutAdlib(0x83+I,0x13);
    PutAdlib(0xE0+I,0x02);
    PutAdlib(0xE3+I,0x00);
    PutAdlib(0xC0+I,0x0A);
    PutAdlib(0xB0+I,0x00);
  }
  PutAdlib(0x34,0x21);
  PutAdlib(0x54,0x3F);
  PutAdlib(0x74,0x99);
  PutAdlib(0x94,0x00);
  PutAdlib(0xF4,0x00);
  PutAdlib(0xBD,0x28);

  return(1);
}

/*** TrashAdlib **************************************/
/*** If Adlib was used then shut it down.          ***/
/*****************************************************/
void TrashAdlib(void)
{
  MuteSound();           /* Turn off the sound      */
  if(adlib_present)
  {
    PutAdlib(0x04,0x60); /* Mask and disable timers */
    PutAdlib(0x04,0x80); /* Reset timers            */
  }
}

void Set_Tone(register UC channel)
{
	register UC J;
	register word freq;
	register long Tune;

	if(channel==3)
	{
		/* FB 'feedback' bit not yet supported */
		/*if(PSG[6]&4)
		{
		}*/
		/* Also I don't know the clock value! */
		switch(PSG[6]&0x03)
		{
			case 0: /* Clock/2 */
				freq=2;break;
			case 1: /* Clock/4 */
				freq=4;break;
			case 2: /* Clock/8 */
				freq=8;break;
			case 3: /* From tone generator 3 */
				freq=PSG[4];break;
		}
	}
	else
	{
		freq=PSG[channel<<1];
	}
	Tune=freq? 2345678L/freq:1;

	for(J=0;Tune>0x0400;Tune>>=1) J++;
	switch(channel)
	{
		case 0:
			PutAdlib(0xA0,Tune&0xFF);
			PutAdlib(0xB0,((Tune>>8)&0x03)|(J<<2)|0x20);
			break;
		case 1:
			PutAdlib(0xA1,Tune&0xFF);
			PutAdlib(0xB1,((Tune>>8)&0x03)|(J<<2)|0x20);
			break;
		case 2:
			PutAdlib(0xA2,Tune&0xFF);
			PutAdlib(0xB2,((Tune>>8)&0x03)|(J<<2)|0x20);
			if((PSG[6]&0x03)==0x03)
			{
				PutAdlib(0xA7,Tune&0xFF);
				PutAdlib(0xB7,((Tune>>8)&0x03)|(J<<2)|0x20);
			}
			break;
		case 3:
			PutAdlib(0xA7,Tune&0xFF);
			PutAdlib(0xB7,((Tune>>8)&0x03)|(J<<2)|0x20);
			break;
	}
}

void Set_Vol(register UC channel)
{
	if(channel<3)
	{
		PutAdlib(0x43+channel,3|(PSG[(channel<<1)|1]<<2));
		/*if((channel==2)&&((PSG[6]&3)==3))
		{
			PutAdlib(0x54,(PSG[(channel<<1)|1]<<2));
		}*/
	}
	else
	{
		if(white_noise) PutAdlib(0x54,3|(PSG[(channel<<1)|1]<<2));
	}
}


/*** ResumeSound *************************************/
/*** Resume sound after MuteSound.                 ***/
/*****************************************************/
void ResumeSound(void)
{
	if(UseAdlib)
	{
		Set_Vol(0);
		Set_Vol(1);
		Set_Vol(2);
		if(white_noise) Set_Vol(3);
	}
}

/*** PSGOut ******************************************/
/*** Write value V into AY8910 register R.         ***/
/*** Appropriate sound routines are called as it   ***/
/*** is done.                                      ***/
/*****************************************************/
void PSGOut(register UC TI_1, register UC TI_2)
{
  register UC channels_used, exit_allowed;
  static UC turn_t,turn_v=0;
  register word freq;
  register unsigned long total_freq;

  if(!all_sound_on) return;
  if(UseAdlib)
  {
    _disable();
    switch((TI_1&0x70)>>4)
    {
	case 0: /* Tone 1 frequency */
		PSG[0]=((TI_2&0x3F)<<4)|(TI_1&0x0F);
		Set_Tone(0);break;
	case 1: /* Tone 1 vol */
		PSG[1]=TI_1&0x0F;
		Set_Vol(0);break;
	case 2: /* Tone 2 frequency */
		PSG[2]=((TI_2&0x3F)<<4)|(TI_1&0x0F);
		Set_Tone(1);break;
	case 3: /* Tone 2 vol */
		PSG[3]=TI_1&0x0F;
		Set_Vol(1);break;
	case 4: /* Tone 3 frequency */
		PSG[4]=((TI_2&0x3F)<<4)|(TI_1&0x0F);
		Set_Tone(2);break;
	case 5: /* Tone 3 vol */
		PSG[5]=TI_1&0x0F;
		Set_Vol(2);break;
	case 6: /* Noise gen control */
		PSG[6]=TI_1&0x07;
		Set_Tone(3);break;
	case 7: /* Noise vol */
		PSG[7]=TI_1&0x0F;
		Set_Vol(3);break;
    }
    _enable();
  }
  else
  {
    switch((TI_1&0x70)>>4)
    {
	case 0: /* Tone 1 frequency */
		PSG[0]=((TI_2&0x3F)<<4)|(TI_1&0x0F);
		break;
	case 1: /* Tone 1 vol */
		PSG[1]=TI_1&0x0F;
		break;
	case 2: /* Tone 2 frequency */
		PSG[2]=((TI_2&0x3F)<<4)|(TI_1&0x0F);
		break;
	case 3: /* Tone 2 vol */
		PSG[3]=TI_1&0x0F;
		break;
	case 4: /* Tone 3 frequency */
		PSG[4]=((TI_2&0x3F)<<4)|(TI_1&0x0F);
		break;
	case 5: /* Tone 3 vol */
		PSG[5]=TI_1&0x0F;
		break;
	case 6: /* Noise gen control */
		PSG[6]=TI_1&0x07;
		break;
	case 7: /* Noise vol */
		PSG[7]=TI_1&0x0F;
		break;
    }

    turn_t=turn_v=0;
    if(PSG[3]<PSG[1]) turn_t=turn_v=1;
    if(PSG[5]<PSG[3]) turn_t=turn_v=2;
    if(PSG[7]<PSG[5])
    {
	turn_v=3;
	if((PSG[6]&3)==3)
	{
		turn_t=2;
	}
	else
	{
		turn_t=3;
	}
    }
    total_freq=PSG[turn_t<<1] ? 117283/PSG[turn_t<<1]:1;
    if(PSG[(turn_v<<1)|1]<12) sound(total_freq); else nosound();
  }
}

void Set_fm_Tone(register UC channel)
{
	register UC J;
	register word freq;
	register long Tune;

	freq=(fm_chip[16|channel]|(fm_chip[16|(channel+3)]<<8));
	Tune=freq? 2345678L/freq:1;

	for(J=0;Tune>0x0400;Tune>>=1) J++;
	PutAdlib(0xA0|channel,Tune&0xFF);
	PutAdlib(0xB0|channel,((Tune>>8)&0x03)|(J<<2)|0x20);

	/*PutAdlib(0xA0+channel,fm_chip[16|channel]);
	PutAdlib(0xB0+channel,fm_chip[16|(channel+3)]);*/

	/*PutAdlib(0xA0+channel,fm_chip[16|channel]);
	freq=fm_chip[32|channel];
	freq=(freq<<10)/1025;
	PutAdlib(0xB0+channel,(freq|0x20));*/

}

void Set_fm_Vol(register UC channel)
{
	PutAdlib(0x43+channel,3|(3<<2));
}

void fm_out(register UC R, register UC V)
{
  register word freq;

  if(UseAdlib)
  {
    if((R&15)>7) return;
    _disable();
    switch(R>>4)
    {
	case 0: break;
	case 1: /* Tone frequency */
		Set_fm_Tone(R&7);break;
	case 2: /*  vol */
		Set_fm_Vol(R&7);break;
	case 3: break;
    }
    _enable();
  }
}
#endif
